/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *  
 *    http://www.apache.org/licenses/LICENSE-2.0
 *  
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License. 
 *  
 */
package org.apache.directory.server.dhcp.service.store;

import com.elitecom.avcontroller.service.impl.ComponentManager;
import com.elitecom.avcontroller.utils.InteAddressUtil;
import com.elitecom.avcontroller.utils.NetgearOUI;
import com.google.common.base.Objects;
import com.google.common.cache.Cache;
import com.google.common.cache.CacheBuilder;
import com.google.common.net.InetAddresses;

import java.io.IOException;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;
import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import org.anarres.dhcp.common.address.InterfaceAddress;
import org.apache.directory.server.dhcp.DhcpException;
import org.apache.directory.server.dhcp.io.DhcpRequestContext;
import org.apache.directory.server.dhcp.messages.DhcpMessage;
import org.apache.directory.server.dhcp.messages.HardwareAddress;
import org.apache.directory.server.dhcp.messages.MessageType;
import org.apache.directory.server.dhcp.options.dhcp.RebindingTimeValue;
import org.apache.directory.server.dhcp.options.dhcp.RenewalTimeValue;
import org.apache.directory.server.dhcp.options.perinterface.BroadcastAddress;
import org.apache.directory.server.dhcp.options.vendor.DomainNameServers;
import org.apache.directory.server.dhcp.options.vendor.Routers;
import org.apache.directory.server.dhcp.service.AbstractDhcpReplyFactory;
import org.apache.directory.server.dhcp.service.manager.AbstractLeaseManager;
import org.apache.directory.server.dhcp.options.vendor.SubnetMask;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Very simple dummy/proof-of-concept implementation of a DhcpStore.
 *
 * @author <a href="mailto:dev@directory.apache.org">Apache Directory Project</a>
 */
public class SimpleStoreLeaseManager extends AbstractLeaseManager {
    // private static final String DEFAULT_INITIAL_CONTEXT_FACTORY =
    // "org.apache.directory.server.core.jndi.CoreContextFactory";
    private final static Logger LOG = LoggerFactory.getLogger(SimpleStoreLeaseManager.class);

    // a map of current leases
    public final List<DhcpConfigSubnet> subnets = new ArrayList<DhcpConfigSubnet>();
    private final Cache<HardwareAddress, Lease> leases = CacheBuilder.newBuilder()
            .expireAfterAccess(TTL_LEASE.maxLeaseTime * 2, TimeUnit.SECONDS)
            .maximumSize(150)
            .recordStats()
            .build();

    private final Cache<String, Lease> ipleases = CacheBuilder.newBuilder()
            .expireAfterAccess(TTL_LEASE.maxLeaseTime * 2, TimeUnit.SECONDS)
            .maximumSize(150)
            .recordStats()
            .build();


    //This will suppress PMD.AvoidUsingHardCodedIP warnings in this class
    @SuppressWarnings("PMD.AvoidUsingHardCodedIP")
    public SimpleStoreLeaseManager() {
//        subnets.add(new DhcpConfigSubnet(
//                NetworkAddress.forString("192.168.8.0/24"),
//                InetAddresses.forString("192.168.8.101"), InetAddresses.forString("192.168.8.179")));
    }



    /**
     * Finds the subnet for the given client address.
     */
    @CheckForNull
    protected DhcpConfigSubnet findSubnet(@Nonnull DhcpRequestContext context) {
        for (InterfaceAddress localAddress : context.getInterfaceAddresses()) {
            for (DhcpConfigSubnet subnet : subnets) {
                if (subnet.contains(localAddress.getAddress()))
                    return subnet;
            }
        }
        return null;
    }

    @Override
    public DhcpMessage leaseOffer(
            DhcpRequestContext context,
            DhcpMessage request,
            InetAddress clientRequestedAddress, long clientRequestedExpirySecs)
            throws DhcpException {
        HardwareAddress hardwareAddress = request.getHardwareAddress();
        if (!NetgearOUI.isNetgearMac(hardwareAddress.toString())){
//            LOG.info("DHCP Discover skip {} ,because it is not netgear oui",hardwareAddress.toString());
            return null;
        }
        Lease lease = leases.getIfPresent(hardwareAddress);
        if (lease != null) {
            lease.setState(Lease.LeaseState.OFFERED);
            return newReply(request, MessageType.DHCPOFFER, lease);
        }

        DhcpConfigSubnet subnet = findSubnet(context);
        if (subnet == null) {
            LOG.info("DHCP Discover skip, subnet can not find !");
            return null;
        }
        long leaseTimeSecs = AbstractDhcpReplyFactory.getLeaseTime(TTL_OFFER, clientRequestedExpirySecs);

        // TODO: Allocate a new address.
        if (clientRequestedAddress==null || !subnet.isInRange(clientRequestedAddress)){
                String ip = allocate(subnet);
                byte[] bytes = InteAddressUtil.ipToByteArray(ip);
                try {
                    clientRequestedAddress = InetAddress.getByAddress(bytes);
                } catch (UnknownHostException e) {
                    e.printStackTrace();
                }

            }

        lease = new Lease(hardwareAddress, clientRequestedAddress);
        lease.setState(Lease.LeaseState.OFFERED);

        lease.setExpires(System.currentTimeMillis() / 1000 + leaseTimeSecs);

        lease.getOptions().setAddressOption(SubnetMask.class, subnet.getNetwork().getNetmaskAddress());
        String ipAddress = ComponentManager.currentSite.getIpAddress();
        lease.getOptions().setAddressListOption(Routers.class, (Inet4Address) context.getInterfaceAddress().getAddress());
        lease.getOptions().setAddressListOption(DomainNameServers.class,(Inet4Address) context.getInterfaceAddress().getAddress());
        lease.getOptions().setAddressOption(BroadcastAddress.class,getBroadcastAddress(subnet));
        lease.getOptions().setIntOption(RebindingTimeValue.class,75600);
        lease.getOptions().setIntOption(RenewalTimeValue.class,43200);

        leases.put(hardwareAddress, lease);
        ipleases.put(lease.getClientAddress().getHostAddress(),lease);
        LOG.info("DHCP allocate " +clientRequestedAddress.getHostAddress() +"to"+hardwareAddress.toString());
        return newReply(request, MessageType.DHCPOFFER, lease);
    }
    private InetAddress getBroadcastAddress(DhcpConfigSubnet subnet){
        String min= subnet.getRangeMin().getHostAddress();
        String prefix=min.substring(0,min.lastIndexOf('.')+1);
        byte[] bytes = InteAddressUtil.ipToByteArray(prefix+"255");
        try {
            InetAddress byAddress = InetAddress.getByAddress(bytes);

            return byAddress;
        } catch (UnknownHostException e) {
            e.printStackTrace();

        }
        return null;
    }
    private  int lastping=0;
    private String allocate(DhcpConfigSubnet subnet){
        String ip="";
       String min= subnet.getRangeMin().getHostAddress();
       String prefix=min.substring(0,min.lastIndexOf('.')+1);
       String minlast=min.substring(min.lastIndexOf('.')+1 );
       if (lastping==0) {
           lastping = Integer.parseInt(minlast);
       }
       while (true)
       {
           if (lastping>=250){
               lastping=100;
           }
           ip=prefix+(lastping);
               try {
                  if (!ipleases.asMap().containsKey(ip)  && !InetAddress.getByName(ip).isReachable(1500)){
                      break;
                  }
               } catch (IOException e) {
                   e.printStackTrace();
               }
               lastping++;
           }
        return ip;
    }
    @Override
    public DhcpMessage leaseRequest(
            DhcpRequestContext context,
            DhcpMessage request,
            InetAddress clientRequestedAddress, long clientRequestedExpirySecs) throws DhcpException {
        HardwareAddress hardwareAddress = request.getHardwareAddress();
        Lease lease = leases.getIfPresent(hardwareAddress);
        if (lease == null)
            return null;
        if (!Objects.equal(lease.getClientAddress(), clientRequestedAddress))
            return null;

        lease.setState(Lease.LeaseState.ACTIVE);
        long leaseTimeSecs = AbstractDhcpReplyFactory.getLeaseTime(TTL_LEASE, clientRequestedExpirySecs);
        lease.setExpires(System.currentTimeMillis() / 1000 + leaseTimeSecs);
        LOG.info("dhcp request for " +clientRequestedAddress.getHostAddress() +"to"+hardwareAddress.toString());
        return newReply(request, MessageType.DHCPACK, lease);
    }

    @Override
    public boolean leaseDecline(
            DhcpRequestContext context,
            DhcpMessage request,
            InetAddress clientAddress) throws DhcpException {
        Lease lease = leases.asMap().remove(request.getHardwareAddress());
        if (lease!=null) {
            ipleases.asMap().remove(lease.getClientAddress().getHostAddress());
        }
        return lease != null
                && Objects.equal(lease.getClientAddress(), clientAddress);
    }

    @Override
    public boolean leaseRelease(
            DhcpRequestContext context,
            DhcpMessage request,
            InetAddress clientAddress) throws DhcpException {
        Lease lease = leases.asMap().remove(request.getHardwareAddress());
        if (lease!=null) {
            ipleases.asMap().remove(lease.getClientAddress().getHostAddress());
        }
        return lease != null
                && Objects.equal(lease.getClientAddress(), clientAddress);
    }
}
